#!/usr/bin/env macruby

framework 'AppKit'
require 'strat.rb'

module Enumerable
    def rest
        self[1..-1]
    end
end

def next_event(app)
	return app.nextEventMatchingMask(NSAnyEventMask, 
                                     untilDate:nil, 
                                     inMode:NSDefaultRunLoopMode, 
                                     dequeue: true)
end

def handle_events
	app = NSApplication.sharedApplication
	while event = next_event(app)
		app.sendEvent(event)
	end
end


class ChoiceBar < NSMatrix
	MAX = 5
	
	def initAt(point)
		initWithFrame(NSMakeRect(point.x, point.y, 1, 1))
		renewRows(1, columns:MAX)
		clear
		self
	end
	
	def setup
		MAX.times do |i|
			putCell(nil, atRow:0, column:i)
		end
		
		@buttons.each_with_index do |button, i|
			putCell(button, atRow:0, column:i)
		end
		
		if @buttons.size == 0
			setFrameSize(NSSize.new(1,1))
            else
			size = @buttons.sort_by{|b| b.cellSize.width}.last.cellSize
			setFrameSize(NSSize.new(size.width * @buttons.size, size.height))
			setCellSize(size)
		end
		setNeedsDisplay(true)
		sv = superview
		sv.setNeedsDisplay(true) unless sv.nil?
	end
	
	def clear
		@buttons = []
		setup
	end
	
	def add(label, &callback)
        button = NSButtonCell.alloc.init
        button.setTitle(label)
        button.setBezelStyle(NSRoundedBezelStyle)
        button.setRepresentedObject(callback)
        button.setTarget(self)
        button.setAction(:clicked)
        @buttons.push(button)
	end
	
	def clicked(sender)
		puts "representedObject"
		callback = selectedCell.representedObject
		callback.call unless callback.nil?
	end
	
	def clicked
		puts "representedObject"
		callback = selectedCell.representedObject
		callback.call unless callback.nil?
	end
	
	def putbs
		@buttons.each {|b| puts "#{b.title}: #{b.action} -> #{b.target}"}
	end
end

class Drawer
	def initialize(terrain, units, highlights)
		@terrain, @units, @highlights = terrain, units, highlights
	end
	
	def terrain(col, row, type); draw(col, row, @terrain[type]); end
	def unit(col, row, type); draw(col, row, @units[type]); end
	def highlight(col, row, type); draw(col, row, @highlights[type]); end
	
	def draw(col, row, tile)
		tile.draw(col, row) unless tile.nil?
	end
end

class ColorTile
	attr_reader :width, :height
	def initialize(width, height, r, g, b, a=1.0)
		@width, @height = width, height
		@color = NSColor.colorWithDeviceRed(r, green:g, blue:b, alpha:a)
	end
	
	def draw (col, row)
		x = col * @width
		y = row * @height
		area = NSMakeRect(x, y, @width, @height)
		@color.set
		NSBezierPath.fillRect(area)
	end
end

class ImageTile
    attr_reader :image, :height, :width
    
    def initialize(filename)
        if ! File.exist?(filename)
            path = NSBundle.mainBundle.resourcePath.fileSystemRepresentation
            filename = File.join(path, filename)
        end
        
        @filename = filename
        @image  = NSImage.alloc.initByReferencingFile(filename)
        @width  = @image.size.width
        @height = @image.size.height
    end
    
    def source
        NSMakeRect(0, 0, @width, @height)
    end
    
    def destination(col, row)
        NSMakeRect(col * @width, row * @height, @width, @height)
    end
    
    def draw(col, row)
        image.drawInRect(destination(col, row),
                         fromRect:source,
                         operation:NSCompositeSourceAtop, 
                         fraction:1.0)
    end
end

class MergedTile
	def initialize(*tiles)
		@tiles = tiles
	end
	
	def draw(col, row)
		@tiles.each{|tile| tile.draw(col, row)}
	end
	
	def height; @tiles.first.height; end
	def width; @tiles.first.width; end
end

CUTE_TERRAIN_SHORTEN_Y = 66
class CuteTile < ImageTile
	def height; @height - CUTE_TERRAIN_SHORTEN_Y; end
	def destination(col, row)
		NSMakeRect(col * @width, row * height, @width, @height)
	end
end

CUTE_TILE_OFFSET_Y = 28

class CuteUnitTile < CuteTile
    def destination(col, row)
        NSMakeRect(col * @width, row * height + CUTE_TILE_OFFSET_Y, @width, @height)
    end
end

class CuteColorTile < ColorTile
    def draw(col, row)
        x = col * @width
        y = row * @height
        area = NSMakeRect(x, y + CUTE_TILE_OFFSET_Y, @width, @height)
        @color.set
        NSBezierPath.fillRect(area)
    end
end

class CuteUnitTile < CuteTile
    def destination(col, row)
        NSMakeRect(col * @width, row * height + CUTE_TILE_OFFSET_Y, @width, @height)
    end
end

class Location
	attr_accessor :col, :row, :terrain, :unit, :highlight, :on_click
	def initialize(col, row, drawer)
		@col, @row, @drawer = col, row, drawer
	end
	
    def draw
    	@drawer.terrain(col, row, terrain) unless terrain.nil?
        @drawer.highlight(col, row, highlight) unless highlight.nil?
        @drawer.unit(col, row, unit) unless unit.nil?
  	end
	
	def clear
		@on_click = nil
		@highlight = nil
	end
end

class TBSView < NSView
	def initWithGrid(grid, responder)
		init
		@grid = grid
		@responder = responder
		self
	end
	
	def mouseDown(event)
		@responder.mouseDown(event)
	end
	
	def drawRect(rect)
		@grid.all_positions.reverse.each do |x, y|
			@grid[x,y].draw
		end
	end
end

class ApplicationGameDelegate < NSObject
	def initWithGame(game)
		init
		@game = game
		self
	end
	
	def applicationDidFinishLaunching(sender)
		@game.run
	end
	
    def applicationWillTerminate(notification)
        puts "Bye!"
        exit
	end
end

class CocoaPlayer < BasePlayer
	DEFAULT_WIDTH = 500
	DEFAULT_HEIGHT = 500
	MESSAGE_HEIGHT = 30
	def cols; self.class.const_get(:COLS); end
	def rows; self.class.const_get(:ROWS); end
	def terrain_tiles; self.class.const_get(:TERRAIN); end
	def unit_tiles; self.class.const_get(:UNITS); end
	def highlight_tiles; self.class.const_get(:HIGHLIGHTS); end
	
	def initialize(name)
		super(name)
		
		@drawer = Drawer.new(terrain_tiles, unit_tiles, highlight_tiles)
		@grid = Matrix.new(cols, rows)
		@grid.all_positions.each do |col, row|
			@grid[col, row] = Location.new(col, row, @drawer)
		end
		
		some_tile = terrain_tiles.values[0]
		@tile_width = some_tile.width
		@tile_height = some_tile.height
		window_width = cols * some_tile.width
		window_height = rows * some_tile.height + MESSAGE_HEIGHT
		puts "first tile: #{some_tile.instance_variables}"
		puts "tile dimensions: #{@tile_width}, #{@tile_height}"
		
		app = NSApplication.sharedApplication
		app.setMainMenu(NSMenu.alloc.init)
		#app.delegate = AppDelegate.new
		
		@view = TBSView.alloc.initWithGrid(@grid, self)
		@window = create_window(window_width, window_height, @view)
		
		messages_y = @view.frame.size.height - MESSAGE_HEIGHT
		@messages = create_messages(messages_y, @view.frame.size.width)
		@view.addSubview(@messages)
        
		@bar = ChoiceBar.alloc.initAt(NSPoint.new(0,0))
		@view.addSubview(@bar)	
		return self
	end
    
	def create_window(width, height, view=nil)
		position = NSMakeRect(0, 0, width, height)
		window = NSWindow.alloc.initWithContentRect([0, 0, width, height],
                                                    styleMask:NSTitledWindowMask + NSClosableWindowMask,
                                                    backing:NSBackingStoreBuffered,
                                                    defer:true)
		window.title      = "Turn Based Strategy"
		window.makeKeyAndOrderFront(nil)
		window.contentView = view
		return window
	end
	
	def create_button_bar(x, y, width)
		label = NSTextFieldCell.alloc.init
		label.stringValue = "Actions: "
		
		button = NSButtonCell.new
		button.title = "Test"
		button.bezelStyle = NSRoundedBezelStyle
		
		height = button.cellSize.height
		
		matrix = NSMatrix.alloc.initWithFrame([x,y,width,height])
		matrix.renewRows(1, columns:2)
		matrix.putCell(label, atRow:0, column:0)
		matrix.putCell(button, atRow:0, column:1)
		matrix.setCellSize(button.cellSize)
		return matrix
	end
	
	def message(text)
		@messages.setString(text)
	end
	
	def create_messages(y, width)
		position = [0,y, width, MESSAGE_HEIGHT]
		messages = NSTextView.alloc.initWithFrame(position)
		return messages
	end
	
	def draw(map)
		map.terrain.all_positions.each do |x,y|
			terrain = map.terrain[x, y]
			terrain_type = terrain.rep.first
			@grid[x, y].terrain = terrain_type
		end
		map.units.all_positions.each do |x, y|
			unit = map.units[x,y]
			unit_type = unit.nil? ? nil :unit.rep.first
			@grid[x, y].unit = unit_type
		end
		@view.needsDisplay = true
	end
	
	def xy_to_colrow(x, y)
		col = (x / @tile_width).to_i
		row = (y / @tile_height).to_i
		return col, row
	end
	
	def mouseDown(event)
		click = event.locationInWindow
		col, row = xy_to_colrow(click.x, click.y)
		if col < cols && row < rows
			callback = @grid[col, row].on_click
			callback.call unless callback.nil?
		end
	end
	
	def on_click(x, y, &callback)
		@grid[x,y].on_click = callback
	end
	
	def do_choose(choices)
        choices.each do |choice|
            choice_rep = choice.rep
            choice_type = choice_rep.first
            ideal = "present_#{choice_type.downcase}_choice"
            if respond_to?(ideal)
                send(ideal, choice_rep, *choice_rep.rest)
                else
                present_choice(choice_rep)
            end
        end
        start_choice
        
        while @choosen_rep.nil?
            handle_events
        end
        
        choosen_rep = @choosen_rep
        @choosen_rep = nil
        end_choice
        
        choices.each do |choice|
            choice_rep = choice.rep
            return yield choice if choice_rep == choosen_rep
        end
        
        raise "Somehow a bad choice was selected"
	end
	
	def make_choice(choosen_rep)
        @choosen_rep = choosen_rep
	end
	
	def start_choice
        @bar.setup
	end
	
	def end_choice
        @bar.clear
        @grid.all_positions.each do |col, row|
            @grid[col, row].clear
        end
	end
	
	def present_choice(rep)
        @bar.add(rep.join(" ")) { make_choice(rep) }
        @bar.putbs
	end
	
	def highlight(x, y, type)
		@grid[x, y].highlight = type
	end
end

class DinoCocoaPlayer < CocoaPlayer
	COLS = 10
	ROWS = 6
	
	TERRAIN = {
    #'Blank' => CuteColorTile.new(W, H, 0.5, 0.5, 0.5),
    #'Grass' => ColorTile.new(W, H, 0.0, 1.0, 0.0),
    'Grass' => CuteTile.new('gfx/Grass Block.png'),
    #'Water' => ColorTile.new(W, H, 0.0, 0.0, 1.0),
    'Water' => CuteTile.new('gfx/Water Block.png'),
    #'Forest' => CuteColorTile.new(W,H, 0.0, 0.8, 0,3),
    'Forest' => MergedTile.new(CuteTile.new('gfx/Grass Block.png'), CuteTile.new('grf/Tree Ugly.png')),
    #'Plains' => ColorTile.new(W,H, 0.8, 0.8, 0.8),
    'Plains' => CuteTile.new('gfx/Stone Block.png'),
	}
	W = TERRAIN.values.first.width
    H = TERRAIN.values.first.height
	BLACK_BLOB = CuteTile.new('gfx/Plain Block.png')
	UNITS = {
    'Soldier' => CuteUnitTile.new('gfx/Character Boy.png'),
    'Doctor' => CuteUnitTile.new('gfx/Character Pink Girl.png'),
    'Ranger' => CuteUnitTile.new('gfx/Character Horn Girl.png'),
    'VRaptor' => CuteUnitTile.new('gfx/Enemy Bug.png'),
    'TRex' => CuteUnitTile.new('gfx/Enemy Bug.png'),
	}
    
	HIGHLIGHTS = {
    'Unit' => CuteColorTile.new(W, H, 0.8, 0.4, 0.0, 0.5),
    'Shoot' => CuteColorTile.new(W, H, 1.0, 0.0, 0.0, 0.5),
    'Move' => CuteColorTile.new(W, H, 0.1, 0.1, 1.0, 0.5),
	}
	
	def present_unit_choice(rep, x, y)
		highlight(x, y, 'Unit')
		on_click(x, y) {make_choice(rep) }
	end
	
	def present_move_choice(rep, x, y)
		highlight(x, y, 'Move')
		on_click(x, y) {make_choice(rep) }
	end
	
	def present_shoot_choice(rep, x, y)
		highlight(x, y, 'Shoot')
		on_click(x, y) {make_choice(rep) }
	end
	
	def present_firstaid_choice(rep, x, y)
		present_shoot_choice(rep, x, y)
	end
	
    def xy_to_colrow(x, y)
        col = (x / @tile_width).to_i
        row = ((y - CUTE_TILE_OFFSET_Y ) / @tile_height).to_i
        return col, row
    end
    
    def create_window(width, height, view=nil)
        super(width, height + CUTE_TILE_OFFSET_Y + 10, view)
    end
end


human = DinoCocoaPlayer.new("Human")
game = run_dino(human)
delegate = ApplicationGameDelegate.alloc.initWithGame(game)
NSApplication.sharedApplication.setDelegate(delegate)
NSApp.run
